# avoid " unresolved reference" from Room
# import

from typing import Optional


class PeerAppData:
    def __init__(self, displayName: str, device: dict, rtpCapabilities: dict = None,
                 sctpCapabilities: dict = None,
                 transports: list = None,
                 producers: list = None, consumers: list = None, dataProducers: list = None,
                 dataConsumers: list = None):
        self._displayName = displayName
        self._device = device
        self._rtpCapabilities = rtpCapabilities
        self._sctpCapabilities = sctpCapabilities
        self._transports = transports
        self._producers = producers
        self._consumers = consumers
        self._dataProducers = dataProducers
        self._dataConsumers = dataConsumers

    @property
    def displayName(self) -> str:
        return self._displayName

    @displayName.setter
    def displayName(self, displayName):
        self._displayName = displayName

    @property
    def device(self) -> dict:
        return self._device

    @property
    def rtpCapabilities(self) -> dict:
        return self._rtpCapabilities

    @property
    def sctpCapabilities(self) -> dict:
        return self._sctpCapabilities

    @property
    def transports(self) -> list:
        return self._transports

    @property
    def producers(self) -> list:
        return self._producers

    @property
    def consumers(self) -> list:
        return self._consumers

    @property
    def dataProducers(self) -> list:
        return self._dataProducers

    @property
    def dataConsumers(self) -> list:
        return self._dataConsumers


class Peer:
    # todo room: Room
    def __init__(self, room, peerId: str = None, data: PeerAppData = None):
        self._peerId = peerId
        self._room = room
        self._data = data

    @property
    def peerId(self) -> str:
        return self._peerId

    @peerId.setter
    def peerId(self, peerId):
        self._peerId = peerId

    @property
    # todo room: -> Room
    def room(self):
        return self._room

    @property
    def data(self) -> PeerAppData:
        return self._data

    @data.setter
    def data(self, data: PeerAppData):
        self._data = data


class Room:
    def __init__(self, serverAddress: str = None, roomId: str = None):
        self._serverAddress = serverAddress
        self._roomId = roomId
        # <peerId, Peer>
        self._peerIdToPeerMap = {}
        # <producerId, Peer>
        # const { peerId } = producer.appData;
        self._producerIdToPeerMap = {}
        # <consumerId, Peer>
        # const { peerId } = consumer.appData;
        self._consumerIdToPeerMap = {}
        # <dataConsumerId, Peer>
        self._dataConsumerIdToPeerMap = {}

    @property
    def serverAddress(self) -> str:
        return self._serverAddress

    @serverAddress.setter
    def serverAddress(self, serverAddress):
        self._serverAddress = serverAddress

    @property
    def roomId(self) -> str:
        return self._roomId

    @roomId.setter
    def roomId(self, roomId):
        self._roomId = roomId

    def getPeerByProducerId(self, producerId: str) -> Peer:
        return self._producerIdToPeerMap[producerId]

    def getPeerByConsumerId(self, consumerId: str) -> Peer:
        return self._consumerIdToPeerMap[consumerId]

    def bindConsumerIdToPeer(self, consumerId: str, peer: Peer):
        self._consumerIdToPeerMap[consumerId] = peer

    def unbindConsumerIdToPeer(self, consumerId: str, peer: Peer):
        del self._consumerIdToPeerMap[consumerId]

    def getPeerByDataConsumerId(self, dataConsumerId: str) -> Peer:
        return self._dataConsumerIdToPeerMap[dataConsumerId]

    def bindDataConsumerIdToPeer(self, consumerId: str, peer: Peer):
        self._dataConsumerIdToPeerMap[consumerId] = peer

    def unbindDataConsumerIdToPeer(self, consumerId: str, peer: Peer):
        del self._dataConsumerIdToPeerMap[consumerId]

    def getPeerByPeerId(self, peerId: str) -> Optional[Peer]:
        if self._peerIdToPeerMap.get(peerId):
            return self._peerIdToPeerMap[peerId]
        else:
            return None

    def addPeer(self, peerId: str, data: PeerAppData) -> Peer:
        newPeer = Peer(room=self, peerId=peerId, data=data)
        self._peerIdToPeerMap[peerId] = newPeer
        return newPeer

    def removePeer(self, peerId: str):
        toRemovePeer = self._peerIdToPeerMap[peerId]
        del self._peerIdToPeerMap[peerId]
        return toRemovePeer



import re
from typing import List
from ...rtp_parameters import RtpEncodingParameters, RTX


def getRtpEncodings(offerMediaDict: dict) -> List[RtpEncodingParameters]:
    ssrcs = set()
    for line in offerMediaDict.get('ssrcs', []):
        ssrc = line.get('id')
        ssrcs.add(ssrc)
    
    if len(ssrcs) == 0:
        raise Exception('no a=ssrc lines found')

    ssrcToRtxSsrc = {}

    # First assume RTX is used.
    for line in offerMediaDict.get('ssrcGroups', []):
        if line.get('semantics') != 'FID':
            continue
        ssrc, rtxSsrc = re.split('\s', line.get('ssrcs'))

        ssrc = int(ssrc)
        rtxSsrc = int(rtxSsrc)

        if ssrc in ssrcs:
            # Remove both the SSRC and RTX SSRC from the set so later we know that they
            # are already handled.
            ssrcs.remove(ssrc)
            ssrcs.remove(rtxSsrc)

            # Add to the map.
            ssrcToRtxSsrc[ssrc] = rtxSsrc
    
    # If the set of SSRCs is not empty it means that RTX is not being used, so take
    # media SSRCs from there.
    for ssrc in ssrcs:
        ssrcToRtxSsrc[ssrc] = None
    
    encodings: List[RtpEncodingParameters] = []
    for ssrc, rtxSsrc in ssrcToRtxSsrc.items():
        encoding = RtpEncodingParameters(ssrc=ssrc)
        if rtxSsrc != None:
            encoding.rtx = RTX(ssrc=rtxSsrc)
        encodings.append(encoding)
    return encodings

# Adds multi-ssrc based simulcast into the given SDP media section offer.
def addLegacySimulcast(offerMediaDict: dict, numStreams: int):
    if numStreams <= 1:
        raise Exception('numStreams must be greater than 1')

    # Get the SSRC.
    ssrcMsidLines = [line for line in offerMediaDict.get('ssrcs', []) if line.get('attribute') == 'msid']
    if not ssrcMsidLines:
        raise Exception('a=ssrc line with msid information not found')
    
    ssrcMsidLine = ssrcMsidLines[0]

    # NOTE: const [ streamId, trackId ] = ssrcMsidLine.value.split(' ')[0];
    streamId, trackId = ssrcMsidLine.get('value').split(' ')
    firstSsrc = ssrcMsidLine.get('id')

    firstRtxSsrc = None

    # Get the SSRC for RTX.
    for line in offerMediaDict.get('ssrcGroups', []):
        if line.get('semantics') != 'FID':
            # False
            continue
        ssrcs = re.split('\s', line.get('ssrcs'))
        if int(ssrcs[0]) == firstSsrc:
            firstRtxSsrc = int(ssrcs[1])
            # True
    
    ssrcCnameLine = [line for line in offerMediaDict.get('ssrcs', []) if line.get('attribute') == 'cname']

    if not ssrcCnameLine:
        raise Exception('a=ssrc line with cname information not found')
    
    cname = ssrcCnameLine.get('value')

    ssrcs = []
    rtxSsrcs = []

    for i in range(numStreams):
        ssrcs.append(firstSsrc + i)
        if firstRtxSsrc != None:
            rtxSsrcs.append(firstRtxSsrc + i)
    
    offerMediaDict['ssrcGroups'] = []
    offerMediaDict['ssrcs'] = []

    offerMediaDict['ssrcGroups'].append({
        'semantics': 'SIM',
        'ssrc': ' '.join(ssrcs)
    })

    for ssrc in ssrcs:
        offerMediaDict['ssrcs'].append({
            'id': ssrc,
            'attribute': 'cname',
            'value': cname
        })
        offerMediaDict['ssrcs'].append({
            'id': ssrc,
            'attribute': 'msid',
            'value': f'{streamId} {trackId}'
        })
    
    for i in range(len(rtxSsrcs)):
        ssrc = ssrcs[i]
        rtxSsrc = rtxSsrcs[i]

        offerMediaDict['ssrc'].append({
            'id': rtxSsrc,
            'attribute': 'cname',
            'value': cname
        })
        offerMediaDict['ssrc'].append({
            'id': rtxSsrc,
            'attribute': 'msid',
            'value': f'{streamId} {trackId}'
        })
        offerMediaDict['ssrc'].append({
            'id': 'FID',
            'ssrcs': f'{ssrc} {rtxSsrc}'
        })




import sys
if sys.version_info >= (3, 8):
    from typing import Optional, Dict, Callable, Literal, List, Any, Union
else:
    from typing import Optional, Dict, Callable, List, Any, Union
    from typing_extensions import Literal

import logging
from pyee import AsyncIOEventEmitter
from aiortc import RTCIceServer
from .handlers.handler_interface import HandlerInterface
from .ortc import ExtendedRtpCapabilities, getExtendedRtpCapabilities, canSend, getRecvRtpCapabilities
from .rtp_parameters import RtpCapabilities
from .sctp_parameters import SctpCapabilities, SctpParameters
from .errors import InvalidStateError
from .transport import InternalTransportOptions, Transport
from .models.transport import IceParameters, IceCandidate, DtlsParameters

class Device:
    def __init__(self, handlerFactory):
        self._observer: AsyncIOEventEmitter = AsyncIOEventEmitter()
        # RTC handler factory.
        self._handlerFactory: Callable[..., HandlerInterface] = handlerFactory
        # Loaded flag.
        self._loaded: bool = False

        # Extended RTP capabilities.
        self._extendedRtpCapabilities: Optional[ExtendedRtpCapabilities] = None
        # Local RTP capabilities for receiving media.
        self._recvRtpCapabilities: Optional[RtpCapabilities] = None
        # Whether we can produce audio/video based on computed extended RTP
        # capabilities.
        self._canProduceByKind: Dict[str, bool] = {
            'audio': False,
            'video': False
        }
        self._sctpCapabilities: Optional[SctpCapabilities] = None

    # The RTC handler name.
    @property
    def handlerName(self) -> str:
        if not self._loaded:
            raise InvalidStateError('not loaded')
        return self._handlerName
    
    # Whether the Device is loaded.
    @property
    def loaded(self) -> bool:
        return self._loaded
    
    # RTP capabilities of the Device for receiving media.
    # @raise {InvalidStateError} if not loaded.
    @property
    def rtpCapabilities(self) -> Optional[RtpCapabilities]:
        if not self._loaded:
            raise InvalidStateError('not loaded')
        return self._recvRtpCapabilities
    
    # SCTP capabilities of the Device.
    # @raise {InvalidStateError} if not loaded.
    @property
    def sctpCapabilities(self) -> Optional[SctpCapabilities]:
        if not self._loaded:
            raise InvalidStateError('not loaded')
        return self._sctpCapabilities
    
    # Observer.
    # @emits newtransport - (Transport)
    @property
    def observer(self):
        return self._observer
    
    # Initialize the Device.
    async def load(self, routerRtpCapabilities: Union[RtpCapabilities, dict]):
        logging.debug(f'Device load() [routerRtpCapabilities:{routerRtpCapabilities}]')
        if isinstance(routerRtpCapabilities, dict):
            routerRtpCapabilities:RtpCapabilities = RtpCapabilities(**routerRtpCapabilities)
        else:
            routerRtpCapabilities:RtpCapabilities = routerRtpCapabilities.copy(deep=True)
        # Temporal handler to get its capabilities.
        if self._loaded:
            logging.warning('already loaded')
            return
        handler: HandlerInterface = self._handlerFactory()
        nativeRtpCapabilities = await handler.getNativeRtpCapabilities()
        logging.debug(f'Device load() | got native RTP capabilities:{nativeRtpCapabilities}')
        # Get extended RTP capabilities.
        self._extendedRtpCapabilities = getExtendedRtpCapabilities(nativeRtpCapabilities, routerRtpCapabilities)
        logging.debug(f'Device load() | got extended RTP capabilities:{self._extendedRtpCapabilities}')
        # Check whether we can produce audio/video.
        self._canProduceByKind['audio'] = canSend('audio', self._extendedRtpCapabilities)
        self._canProduceByKind['video'] = canSend('video', self._extendedRtpCapabilities)
        # Generate our receiving RTP capabilities for receiving media.
        self._recvRtpCapabilities = getRecvRtpCapabilities(self._extendedRtpCapabilities)
        logging.debug(f'Device load() | got receiving RTP capabilities:{self._recvRtpCapabilities}')
        # Generate our SCTP capabilities.
        self._sctpCapabilities = await handler.getNativeSctpCapabilities()
        logging.debug(f'Device load() | got native SCTP capabilities:{self._sctpCapabilities}')
        logging.debug('Device load() succeeded')
        self._loaded = True
        self._handlerName = handler.name
        await handler.close()

    # Whether we can produce audio/video.
    # @raise {InvalidStateError} if not loaded.
    # @raise {TypeError} if wrong arguments.
    def canProduce(self, kind: Literal['video', 'audio']):
        if not self._loaded:
            raise InvalidStateError('not loaded')
        elif kind not in ['video', 'audio']:
            raise TypeError(f'invalid kind {kind}')
        return self._canProduceByKind[kind]
    
    # Creates a Transport for sending media.
    # @raise {InvalidStateError} if not loaded.
    # @raise {TypeError} if wrong arguments.
    def createSendTransport(
        self,
        id: str,
        iceParameters: IceParameters,
        iceCandidates: List[IceCandidate],
        dtlsParameters: DtlsParameters,
        sctpParameters: Optional[SctpParameters],
        iceServers: Optional[List[RTCIceServer]] = None,
        iceTransportPolicy: Optional[Literal['all', 'relay']] = None,
        additionalSettings: Optional[dict] = None,
        proprietaryConstraints: Any = None,
        appData: Optional[dict] = {}
    ) -> Transport:
        logging.debug('createSendTransport()')
        return self._createTransport(
            direction='send',
            id=id,
            iceParameters=iceParameters,
            iceCandidates=iceCandidates,
            dtlsParameters=dtlsParameters,
            sctpParameters=sctpParameters,
            iceServers=iceServers,
            iceTransportPolicy=iceTransportPolicy,
            additionalSettings=additionalSettings,
            proprietaryConstraints=proprietaryConstraints,
            appData=appData
        )
    
    # Creates a Transport for receiving media.
    # @raise {InvalidStateError} if not loaded.
    # @raise {TypeError} if wrong arguments.
    def createRecvTransport(
        self,
        id: str,
        iceParameters: Union[IceParameters, dict],
        iceCandidates: List[Union[IceCandidate, dict]],
        dtlsParameters: Union[DtlsParameters, dict],
        sctpParameters: Optional[Union[SctpParameters, dict]] = None,
        iceServers: Optional[List[RTCIceServer]] = None,
        iceTransportPolicy: Optional[Literal['all', 'relay']] = None,
        additionalSettings: Optional[dict] = None,
        proprietaryConstraints: Any = None,
        appData: Optional[dict] = {}
    ) -> Transport:
        logging.debug('createRecvTransport()')
        if isinstance(iceParameters, dict):
            iceParameters: IceParameters = IceParameters(**iceParameters)
        
        iceCandidates = [IceCandidate(**c) if isinstance(c, dict) else c for c in iceCandidates]
        
        if isinstance(dtlsParameters, dict):
            dtlsParameters: DtlsParameters = DtlsParameters(**dtlsParameters)
        
        if isinstance(sctpParameters, dict):
            sctpParameters: SctpParameters = SctpParameters(**sctpParameters)
        
        return self._createTransport(
            direction='recv',
            id=id,
            iceParameters=iceParameters,
            iceCandidates=iceCandidates,
            dtlsParameters=dtlsParameters,
            sctpParameters=sctpParameters,
            iceServers=iceServers,
            iceTransportPolicy=iceTransportPolicy,
            additionalSettings=additionalSettings,
            proprietaryConstraints=proprietaryConstraints,
            appData=appData
        )
    
    def _createTransport(
        self,
        direction: Literal['send', 'recv'],
        id: str,
        iceParameters: IceParameters,
        iceCandidates: List[IceCandidate],
        dtlsParameters: DtlsParameters,
        sctpParameters: Optional[SctpParameters],
        iceServers: Optional[List[RTCIceServer]] = None,
        iceTransportPolicy: Optional[Literal['all', 'relay']] = None,
        additionalSettings: Optional[dict] = None,
        proprietaryConstraints: Any = None,
        appData: Optional[dict] = {}
    ) -> Transport:
        if not self._loaded:
            raise InvalidStateError('not loaded')
        transport = Transport(options=InternalTransportOptions(
            direction=direction,
            handlerFactory=self._handlerFactory,
            extendedRtpCapabilities=self._extendedRtpCapabilities,
            canProduceByKind=self._canProduceByKind,
            id=id,
            iceParameters=iceParameters,
            iceCandidates=iceCandidates,
            dtlsParameters=dtlsParameters,
            sctpParameters=sctpParameters,
            iceServers=iceServers,
            iceTransportPolicy=iceTransportPolicy,
            additionalSettings=additionalSettings,
            proprietaryConstraints=proprietaryConstraints,
            appData=appData
        ))
    
        return transport
